#include <stdio.h>
#include <errno.h>
#include <math.h>

#include "DEFS.h"
#include "PROB.h"

#ifdef BUNDLE_AR_WITH_PROB

#include "RNG.h"

int GaussianT 
            (
            int NumRates, 
            REAL *shift,
            REAL *RatesPres, 
            REAL *RatesPrev, 
            REAL *RatesNext, 
            REAL **Couplings
            )

{
   REAL arg;
   REAL Delta;
   int Naccepted = 0;
   int TrialIndex;
   int RateIndex;
   REAL u;

   for(TrialIndex=1;TrialIndex<=NumRates;TrialIndex++) {
     u = shift[TrialIndex];
     arg=0.0;
     for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
       Delta  = RatesPres[RateIndex]-RatesPrev[RateIndex];
       Delta *= Couplings[TrialIndex][RateIndex];
       arg +=  Delta;
     }
     arg *= u;
     arg += 0.5*u*u*Couplings[TrialIndex][TrialIndex];
     arg = -arg;
  
     arg = (arg > MAXARG) ? MAXARG:arg;
     arg = (arg < MINARG) ? MINARG:arg;

     if(randf() <= exp(arg)) {
         Naccepted++;
         RatesPres[TrialIndex] += u;
     } else {
         shift[TrialIndex] = 0.0;
     }
   }
   return Naccepted;
}


int GaussianTEV
              (
              int NumRates,
              REAL *shift,
              REAL *RatesPres,
              REAL *RatesPrev,
              REAL *RatesNext,
              REAL **Couplings
              )

{
   REAL arg;
   int Naccepted = 0;
   int RateIndex;
   REAL u;
   REAL c;

   for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
     u = shift[RateIndex];
     c = Couplings[RateIndex][1];
     arg  = u*c*(RatesPres[RateIndex]-RatesPrev[RateIndex]);
     arg += 0.5*u*u*c;
     arg = -arg;

     arg = (arg > MAXARG) ? MAXARG:arg;
     arg = (arg < MINARG) ? MINARG:arg;

     if(randf() <= exp(arg)) {
         Naccepted++;
     } else {
         shift[RateIndex] = 0.0;
     }
   }
   return Naccepted;
}

int Gaussian 
           (
           int NumRates, 
           REAL *shift,
           REAL *RatesPres, 
           REAL *RatesPrev, 
           REAL *RatesNext, 
           REAL **Couplings
           )

{
   REAL arg;
   REAL Delta;
   int RateIndex;
   int TrialIndex;
   int Naccepted = 0;
   REAL arg1;
   REAL arg2;
   REAL u;
   REAL v;

   if(RatesNext != NULL) {

      for(TrialIndex=1;TrialIndex<=NumRates;TrialIndex++) {

        arg1=0.0;
        u = shift[TrialIndex];
        v = 0.5*u*u*Couplings[TrialIndex][TrialIndex];

        for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
          Delta  = RatesPres[RateIndex]-RatesPrev[RateIndex];
          Delta *= Couplings[TrialIndex][RateIndex];
          arg1 +=  Delta;
        }
        arg1 *= u;
        arg1 += v;
        arg1 = -arg1;
   
        arg2=0.0;
        for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
          Delta  = RatesNext[RateIndex]-RatesPres[RateIndex];
          Delta *= Couplings[TrialIndex][RateIndex];
          arg2 +=  Delta;
        }
        arg2 *= (-1.)*u;
        arg2 += v;
        arg2 = -arg2;
   
        arg = arg1 + arg2;
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;

        if(randf() <= exp(arg)) {
            Naccepted++;
            RatesPres[TrialIndex] += u;
        } else {
            shift[TrialIndex] = 0.0;
        }

      }
   
   }  else  {

      for(TrialIndex=1;TrialIndex<=NumRates;TrialIndex++) {

        arg=0.0;
        u = shift[TrialIndex];

        for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
          Delta  = RatesPres[RateIndex]-RatesPrev[RateIndex];
          Delta *= Couplings[TrialIndex][RateIndex];
          arg +=  Delta;
        }
        arg *= u;
        arg += 0.5*u*u*Couplings[TrialIndex][TrialIndex];
        arg = -arg;
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;

        if(randf() <= exp(arg)) {
            Naccepted++;
            RatesPres[TrialIndex] += u;
        } else {
            shift[TrialIndex] = 0.0;
        }

      }
  }

  return Naccepted;
}


int Student3T
            (
            int NumRates,
            REAL *update,
            REAL *RatesPres,
            REAL *RatesPrev,
            REAL *RatesNext,
            REAL **Couplings
            )

{
   int RateIndex;
   int TrialIndex;
   int ir;
   int jr;
   int Naccepted = 0;

   REAL arg1;
   REAL arg;
   REAL term;
   REAL prob_fi1;
   REAL prob_fi;
   REAL arg_i1;
   REAL arg_i;
   REAL shift;
   REAL delta;

   REAL v1;
   REAL v2;
   const int v=3;

   v1 = (NumRates*v+1.)/2.;
   v2 = (REAL) (v*NumRates);

   arg_i = 0.0;
   for(ir=1;ir <= NumRates;ir++) for(jr=1;jr <=NumRates;jr++) {
     term = Couplings[ir][jr]*(RatesPres[ir]-RatesPrev[ir]) * 
                              (RatesPres[jr]-RatesPrev[jr]);
     arg_i += term;
   } /* initial value of quadratic form  */

   for(TrialIndex=1; TrialIndex <= NumRates; TrialIndex++) {

     shift = update[TrialIndex];
     arg_i1=arg_i;
     arg1=arg_i;

     /* terms linear in shift */
     delta = 0.0;
     term = 0.0;
     for(jr=1;jr <=NumRates;jr++) {
       term += (RatesPres[jr]-RatesPrev[jr])*Couplings[TrialIndex][jr];
     }
     term *= 2.*shift;
     delta += term;
     arg1 += term;

     /* term quadratic in shift */
     delta += (term=shift*shift*Couplings[TrialIndex][TrialIndex]);
     arg1 += term;

     arg_i1 /= v2;
     arg1 /= v2;

     prob_fi1= (1. + arg_i1 )/(1. + arg1);
     arg = v1*log(prob_fi1);
     arg = (arg > MAXARG) ? MAXARG:arg;
     arg = (arg < MINARG) ? MINARG:arg;

     prob_fi = exp(arg);
     if(randf() <= exp(arg)) {
         Naccepted++;
         arg_i += delta;
         RatesPres[TrialIndex] += shift;
     } else {
         update[TrialIndex] = 0.0;
     }


   } /* end of loop over rates */

   return Naccepted;

}/* end of generic Student (n=3) with translation symmetry */



int GeneralizedCauchy
                    (
                    int NumRates,
                    REAL *update,
                    REAL *RatesPres,
                    REAL *RatesPrev,
                    REAL *RatesNext,
                    REAL **Parameter
                    )

{
   int Naccepted = 0;

   int rIndex;
   int ir;
   int jr;

   REAL arg;
   REAL term;
   REAL arg_i;
   REAL shift;
   REAL delta;
   REAL *Rates = NULL;

   Rates =  (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchy()");

   if(RatesNext == NULL) {

      for(ir=1;ir <= NumRates;ir++) {
        Rates[ir] = RatesPres[ir] - RatesPrev[ir];
      }

/*    initial value of quadratic form  */
      arg_i = 0.0;
      for(ir=1;ir <= NumRates;ir++) {
        for(jr=1;jr <=NumRates;jr++) {
             arg_i += Parameter[ir][jr]*Rates[ir]*Rates[jr];
        }
      }
         
/******
      (1) loop over rates,
      (2) apply trial update to each rate
      (3) compute ratio of probabilities
      (4) accept/reject
      (5) update quadratic form
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {

        shift = update[rIndex];

        /* terms linear in shift */
        delta = 0.0;
        term = 0.0;
        for(jr=1;jr <=NumRates;jr++) {
          term += Rates[jr]*Parameter[rIndex][jr];
        }
        term *= 2.*shift;
        delta += term;
   
        /* term quadratic in shift */
        delta += (term=shift*shift*Parameter[rIndex][rIndex]);
   
        arg = (1. + arg_i )/(1. + arg_i + delta);
        arg = Parameter[1][NumRates+1]*log(arg);
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;
   
        if(randf() <= exp(arg) ){
          Naccepted++;
          Rates[rIndex] += shift;
          arg_i += delta;
        } else {
          update[rIndex] = 0.0;
        }
   
     } /* end of loop over rates */


  } else {

   REAL arg1;
   REAL term1;
   REAL arg_i_1;
   REAL shift1;
   REAL delta1;
   REAL *Rates1 = NULL;
   REAL arg2;
   REAL term2;
   REAL arg_i_2;
   REAL shift2;
   REAL delta2;
   REAL *Rates2=NULL;

      Rates1 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchy()");
      Rates2 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchy()");

      for(ir=1;ir <= NumRates;ir++) {
        Rates1[ir] = RatesPres[ir] - RatesPrev[ir];
        Rates2[ir] = RatesNext[ir] - RatesPres[ir];
      }

/*    initial values of quadratic forms on next and prev time steps  */
      arg_i_1 = 0.0;
      arg_i_2 = 0.0;
      for(ir=1;ir <= NumRates;ir++) {
        for(jr=1;jr <=NumRates;jr++) {
             arg_i_1 += Parameter[ir][jr]*Rates1[ir]*Rates1[jr];
             arg_i_2 += Parameter[ir][jr]*Rates2[ir]*Rates2[jr];
        }
      }
        
/******
      (1) loop over rates,
      (2) apply trial update to each rate
      (3) compute ratio of probabilities; probability is a product
          of next and prev time steps
      (4) accept/reject
      (5) update quadratic form
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {

        shift1 = update[rIndex];
        shift2 = -update[rIndex];

        /* terms linear in shift */
        delta1 = 0.0;
        delta2 = 0.0;
        term1 = 0.0;
        term2 = 0.0;
        for(jr=1;jr <=NumRates;jr++) {
          term1 += Rates1[jr]*Parameter[rIndex][jr];
          term2 += Rates2[jr]*Parameter[rIndex][jr];
        }
        term1 *= 2.*shift1;
        term2 *= 2.*shift2;
        delta1 += term1;
        delta2 += term2;
   
        /* term quadratic in shift */
        delta1 += (term1=shift1*shift1*Parameter[rIndex][rIndex]);
        delta2 += (term2=shift2*shift2*Parameter[rIndex][rIndex]);
  
        arg1 = (1. + arg_i_1 )/(1. + arg_i_1 + delta1);
        arg2 = (1. + arg_i_2 )/(1. + arg_i_2 + delta2);
        arg1 = Parameter[1][NumRates+1]*log(arg1);
        arg2 = Parameter[1][NumRates+1]*log(arg2);
        arg  = arg1 + arg2;
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;
  
        if(randf() <= exp(arg)){
          Naccepted++;
          Rates[rIndex] += shift1;
          arg_i_1 += delta1;
          arg_i_2 += delta2;
        } else {
          update[rIndex] = 0.0;
        }
  
     } /* end of loop over rates */

  }

  return Naccepted;

} /* end of GeneralizedCauchy() */



int GeneralizedCauchyEV
                      (
                      int NumRates,
                      REAL *update,
                      REAL *RatesPres,
                      REAL *RatesPrev,
                      REAL *RatesNext,
                      REAL **Parameter
                      )

{
   int Naccepted = 0;

   int rIndex;
   int ir;

   REAL arg;
   REAL term;
   REAL arg_i;
   REAL shift;
   REAL delta;
   REAL *Rates = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchyEV()");

   REAL arg1;
   REAL term1;
   REAL arg_i_1;
   REAL shift1;
   REAL delta1;
   REAL *Rates1 = NULL;

   REAL arg2;
   REAL term2;
   REAL arg_i_2;
   REAL shift2;
   REAL delta2;
   REAL *Rates2=NULL;

   Rates1 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchyEV()");
   Rates2 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedCauchyEV()");

   if(RatesNext == NULL) {

      for(ir=1;ir <= NumRates;ir++) {
        Rates[ir] = RatesPres[ir] - RatesPrev[ir];
      }

         
/******
      (1) loop over rates,
      (2) apply trial update to each rate
      (3) compute ratio of probabilities
      (4) accept/reject
      (5) update quadratic form
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {

        shift = update[rIndex];
/* Scott found a bug here.....
        arg_i = Parameter[ir][1]*Rates[ir]*Rates[ir];
ir should be rIndex
*/
        arg_i = Parameter[rIndex][1]*Rates[rIndex]*Rates[rIndex];

        /* term linear in shift */
        delta = 0.0;
        term = 2.*shift*Rates[rIndex]*Parameter[rIndex][1];
        delta += term;
   
        /* term quadratic in shift */
        delta += (term=shift*shift*Parameter[rIndex][1]);
   
        arg = (1. + arg_i )/(1. + arg_i + delta);
/* Scott found a bug here.....
        arg = Parameter[rIndex][2]*log(arg);
here we address index 2, and I believe that it should be index 1


*/

        arg = Parameter[rIndex][2]*log(arg);
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;
   
        if(randf() <= exp(arg) ){
          Naccepted++;
          Rates[rIndex] += shift;
        } else {
          update[rIndex] = 0.0;
        }
   
     } /* end of loop over rates */


  } else {

      for(ir=1;ir <= NumRates;ir++) {
        Rates1[ir] = RatesPres[ir] - RatesPrev[ir];
        Rates2[ir] = RatesNext[ir] - RatesPres[ir];
      }

/******
      (1) loop over rates,
      (2) apply trial update to each rate
      (3) compute ratio of probabilities; probability is a product
          of next and prev time steps
      (4) accept/reject
      (5) update quadratic form
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {

        shift1 = update[rIndex];
        shift2 = -update[rIndex];
/* Scott found a bug here.....
   I believe that instead of the following lines

        arg_i_1 = Parameter[ir][1]*Rates1[ir]*Rates1[ir];
        arg_i_2 = Parameter[ir][1]*Rates2[ir]*Rates2[ir];
   we should have

        arg_i_1 = Parameter[rIndex][1]*Rates1[rIndex]*Rates1[rIndex];
        arg_i_2 = Parameter[rIndex][1]*Rates2[rIndex]*Rates2[rIndex];
*/
        arg_i_1 = Parameter[rIndex][1]*Rates1[rIndex]*Rates1[rIndex];
        arg_i_2 = Parameter[rIndex][1]*Rates2[rIndex]*Rates2[rIndex];

        /* terms linear in shift */
        delta1 = 0.0;
        delta2 = 0.0;
        term1 =2.*shift1*Rates1[rIndex]*Parameter[rIndex][1];
        term2 =2.*shift2*Rates2[rIndex]*Parameter[rIndex][1];
        delta1 += term1;
        delta2 += term2;
   
        /* term quadratic in shift */
        delta1 += (term1=shift1*shift1*Parameter[rIndex][1]);
        delta2 += (term2=shift2*shift2*Parameter[rIndex][1]);
  
        arg1 = (1. + arg_i_1 )/(1. + arg_i_1 + delta1);
        arg2 = (1. + arg_i_2 )/(1. + arg_i_2 + delta2);
/* Scott found a bug here.....
        arg1 = Parameter[rIndex][2]*log(arg1);
        arg2 = Parameter[rIndex][2]*log(arg2);
these 2 lines have a 2 for the second index.... the array is only allocated
for 1 (0:1), so this must be a problem....
I will put a 1 here....

*/
        arg1 = Parameter[rIndex][2]*log(arg1);
        arg2 = Parameter[rIndex][2]*log(arg2);
        arg  = arg1 + arg2;
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;
  
        if(randf() <= exp(arg)){
          Naccepted++;
          Rates[rIndex] += shift1;
        } else {
          update[rIndex] = 0.0;
        }
  
     } /* end of loop over rates */

  }

  V_Free((void *) Rates);
  V_Free((void *) Rates1);
  V_Free((void *) Rates2);

  return Naccepted;

} /* end of GeneralizedCauchyEV() */




int GeneralizedLogistic
                      (
                      int NumRates,
                      REAL *update,
                      REAL *RatesPres,
                      REAL *RatesPrev,
                      REAL *RatesNext,
                      REAL **Parameter
                      )

{
   int Naccepted = 0;

   int rIndex;
   int ir;

   REAL arg;
   REAL arg1;
   REAL arg2;
   REAL arg3;
   REAL arg_i;
   REAL t;
   REAL v;
   REAL shift;
   REAL delta;

/* the exponent of the denominator */
      v = Parameter[ir=1][4];
      for(ir=1;ir <=  NumRates; ir++) v += Parameter[ir][1];

   if(RatesNext == NULL) {

   REAL *term = (REAL *) mymalloc((NumRates+1)*sizeof(REAL), "GeneralizedLogistic()");
   REAL *Rates =  (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogistic()");

      for(ir=1;ir <= NumRates;ir++) {
        Rates[ir] = RatesPres[ir] - RatesPrev[ir];
      }

/* initial value of the sum of exponentials in denominator */
      arg_i = 0.0;
      for(ir=1;ir <= NumRates;ir++) {
        t = Parameter[ir][2]*Rates[ir]-Parameter[ir][3];
        term[ir] = t = exp(-t*Parameter[ir][1]);
        arg_i += t;
      }

/******
   (1) loop over rates,
   (2) apply trial update to each rate
   (3) compute ratio of probabilities
   (4) accept/reject
   (5) update value of the sum of exponentials in denominator
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {
   
        shift = update[rIndex];
        arg2=arg1=1.+arg_i;
        arg2 += (delta =  term[rIndex] * (exp(-Parameter[rIndex][2]*shift) -1.));
        arg3 = Parameter[rIndex][1]*Parameter[rIndex][2]*shift;
   
        arg = arg1 / arg2;
        arg = -arg3 + v*log(arg);
        arg = (arg > MAXARG) ? MAXARG:arg;
        arg = (arg < MINARG) ? MINARG:arg;
   
        if(randf() <= exp(arg) ){
          Naccepted++;
          arg_i += delta;
        } else {
          update[rIndex] = 0.0;
        }
   
     } /* end of loop over rates */
     V_Free((void *) term);
     V_Free((void *) Rates);



  } else {

   REAL arg_i_1;
   REAL arg_i_2;
   REAL arg_1;
   REAL arg_2;
   REAL arg1_1;
   REAL arg1_2;
   REAL arg2_1;
   REAL arg2_2;
   REAL arg3_1;
   REAL arg3_2;
   REAL shift1;
   REAL shift2;
   REAL t1;
   REAL t2;
   REAL delta1;
   REAL delta2;

   REAL *Rates1 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogistic()");
   REAL *Rates2 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogistic()");
   REAL *term1 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogistic()");
   REAL *term2 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogistic()");

     for(ir=1;ir <= NumRates;ir++) {
        Rates1[ir] = RatesPres[ir] - RatesPrev[ir];
        Rates2[ir] = RatesNext[ir] - RatesPres[ir];
     }

/* initial value of the sum of exponentials in denominator */
     arg_i_1 = 0.0;
     arg_i_2 = 0.0;

     for(ir=1;ir <= NumRates;ir++) {
       t1 = Parameter[ir][2]*Rates1[ir]-Parameter[ir][3];
       t2 = Parameter[ir][2]*Rates2[ir]-Parameter[ir][3];
       term1[ir] = t1 = exp(-t1*Parameter[ir][1]);
       term2[ir] = t2 = exp(-t2*Parameter[ir][1]);
       arg_i_1 += t1;
       arg_i_2 += t2;
     }
  
/******
   (1) loop over rates,
   (2) apply trial update to each rate
   (3) compute ratio of probabilities
   (4) accept/reject
   (5) update value of the sum of exponentials in denominator
 *****/

     for(rIndex=1; rIndex <= NumRates; rIndex++) {
  
       shift1 = update[rIndex];
       shift2 = -update[rIndex];
       arg2_1=arg1_1=1.+arg_i_1;
       arg2_2=arg1_2=1.+arg_i_2;
       arg2_1 += (delta1 =  term1[rIndex] * (exp(-Parameter[rIndex][2]*shift1) -1.));
       arg2_2 += (delta2 =  term2[rIndex] * (exp(-Parameter[rIndex][2]*shift2) -1.));
       arg3_1 = Parameter[rIndex][1]*Parameter[rIndex][2]*shift1;
       arg3_2 = Parameter[rIndex][1]*Parameter[rIndex][2]*shift2;
  
       arg_1 = arg1_1 / arg2_1;
       arg_2 = arg1_2 / arg2_2;
       arg_1 = -arg3_1 + v*log(arg_1);
       arg_2 = -arg3_2 + v*log(arg_2);

       arg = arg_1 + arg_2;
       arg = (arg > MAXARG) ? MAXARG:arg;
       arg = (arg < MINARG) ? MINARG:arg;
  
       if(randf() <= exp(arg) ){
         Naccepted++;
         arg_i_1 += delta1;
         arg_i_2 += delta2;
       } else {
          update[rIndex] = 0.0;
       }
  
    } /* end of loop over rates */

    V_Free((void *) term1);
    V_Free((void *) term2);
    V_Free((void *) Rates1);
    V_Free((void *) Rates2);

  } /* end of if(Next==NULL)-else */

  return Naccepted;

} /* end of GeneralizedLogistic() */



int GeneralizedLogisticEV
                        (
                        int NumRates,
                        REAL *update,
                        REAL *RatesPres,
                        REAL *RatesPrev,
                        REAL *RatesNext,
                        REAL **Parameter
                        )

{
   int Naccepted = 0;
   int rIndex;
   int ir;
   REAL fac;
   REAL arg;


   if(RatesNext == NULL) {

   REAL *Rates = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogisticEV()");
   REAL arg1;
   REAL arg2;
   REAL arg3;
   REAL arg_i;
   REAL shift;

      for(ir=1;ir <= NumRates;ir++) {
        Rates[ir] = RatesPres[ir] - RatesPrev[ir];
      }

/******
   (1) loop over rates,
   (2) apply trial update to each rate
   (3) compute ratio of probabilities
   (4) accept/reject
 *****/

      for(rIndex=1; rIndex <= NumRates; rIndex++) {
  
       fac = Parameter[rIndex][1];
       shift = update[rIndex];
       arg1 = exp(
                  -(fac*Rates[rIndex]-Parameter[rIndex][2])
                 );
       arg3 = fac*shift;
       arg2 = 1.+arg1*arg3;
       arg1 += 1.;
  
  
       arg = arg1 / arg2;
       arg = -arg3 + Parameter[rIndex][3]*log(arg);
       arg = (arg > MAXARG) ? MAXARG:arg;
       arg = (arg < MINARG) ? MINARG:arg;
  
       if(randf() <= exp(arg) ){
         Naccepted++;
       } else {
         update[rIndex] = 0.0;
       }
  
    } /* end of loop over rates */

    V_Free((void *) Rates);


  } else {

   REAL arg_1;
   REAL arg_2;
   REAL arg1_1;
   REAL arg1_2;
   REAL arg2_1;
   REAL arg2_2;
   REAL arg3_1;
   REAL arg3_2;
   REAL arg_i_1;
   REAL arg_i_2;
   REAL shift1;
   REAL shift2;


   REAL *Rates1 = NULL;
   REAL *Rates2 = NULL;

      Rates1 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogisticEV()");
      Rates2 = (REAL *) mymalloc( (NumRates+1) *sizeof(REAL),"GeneralizedLogisticEV()");

      for(ir=1;ir <= NumRates;ir++) {
        Rates1[ir] = RatesPres[ir] - RatesPrev[ir];
        Rates2[ir] = RatesNext[ir] - RatesPres[ir];
      }

      for(rIndex=1; rIndex <= NumRates; rIndex++) {
 
       fac = Parameter[rIndex][1];
       shift1 = update[rIndex];
       shift2 = -update[rIndex];
       arg1_1 = exp(
                   -(fac*Rates1[rIndex]-Parameter[rIndex][2])
                   );
       arg1_2 = exp(
                   -(fac*Rates2[rIndex]-Parameter[rIndex][2])
                   );
       arg3_1 = fac*shift1;
       arg3_2 = fac*shift2;
       arg2_1 = 1.+arg1_1*arg3_1;
       arg2_2 = 1.+arg1_2*arg3_2;
       arg1_1 += 1.;
       arg1_2 += 1.;
 
 
       arg_1 = arg1_1 / arg2_1;
       arg_2 = arg1_2 / arg2_2;
       arg_1 = -arg3_1 + Parameter[rIndex][3]*log(arg_1);
       arg_2 = -arg3_2 + Parameter[rIndex][3]*log(arg_2);
       arg =  arg_1 + arg_2;
       arg = (arg > MAXARG) ? MAXARG:arg;
       arg = (arg < MINARG) ? MINARG:arg;

       if(randf() <= exp(arg) ){
         Naccepted++;
       } else {
         update[rIndex] = 0.0;
       }

    } /* end of loop over rates */

    V_Free((void *) Rates1);
    V_Free((void *) Rates2);


  }

  return Naccepted;

} /* end of GeneralizedLogisticEV() */



#else


REAL GaussianT 
              (
              int NumRates, 
              int TrialIndex, 
              REAL shift,
              REAL *RatesPres, 
              REAL *RatesPrev, 
              REAL *RatesNext, 
              REAL **Couplings
              )

{
   REAL arg;
   REAL Delta;
   int RateIndex;

   arg=0.0;
   for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
     Delta  = RatesPres[RateIndex]-RatesPrev[RateIndex];
     Delta *= Couplings[TrialIndex][RateIndex];
     arg +=  Delta;
   }
   arg *= shift;
   arg += 0.5*shift*shift*Couplings[TrialIndex][TrialIndex];
   arg = -arg;

   arg = (arg > MAXARG) ? MAXARG:arg;
   arg = (arg < MINARG) ? MINARG:arg;

   return exp(arg);
}

REAL Gaussian 
             (
             int NumRates, 
             int TrialIndex, 
             REAL shift,
             REAL *RatesPres, 
             REAL *RatesPrev, 
             REAL *RatesNext, 
             REAL **Couplings
             )

{
   REAL arg;
   REAL Delta;
   int RateIndex;
   REAL arg1; 
   REAL arg2;
   REAL v;

   v = 0.5*shift*shift*Couplings[TrialIndex][TrialIndex];
   arg1=0.0;

   for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
     Delta  = RatesPres[RateIndex]-RatesPrev[RateIndex];
     Delta *= Couplings[TrialIndex][RateIndex];
     arg1 +=  Delta;
   }
   arg1 *= shift;
   arg1 += v;
   arg1 = -arg1;

   arg2=0.0;
   if(RatesNext != NULL) {
     for(RateIndex=1;RateIndex<=NumRates;RateIndex++) {
       Delta  = RatesNext[RateIndex]-RatesPres[RateIndex];
       Delta *= Couplings[TrialIndex][RateIndex];
       arg2 +=  Delta;
     }
     arg2 *= (-1.)*shift;
     arg2 += v;
     arg2 = -arg2;
   }

   arg = arg1 + arg2;

   arg = (arg > MAXARG) ? MAXARG:arg;
   arg = (arg < MINARG) ? MINARG:arg;

   return exp(arg);
}


REAL Student3T
              (
              int NumRates,
              int TrialIndex,
              REAL shift,
              REAL *RatesPres,
              REAL *RatesPrev,
              REAL *RatesNext,
              REAL **Couplings
              )

{
   int RateIndex;
   int ir;
   int jr;

   REAL arg1;
   REAL arg;
   REAL term;
   REAL prob_fi1;
   REAL prob_fi;
   REAL arg_i1;

   REAL v1;
   REAL v2;
   const int v=3;


   v1 = (NumRates*v+1.)/2.;
   v2 = (REAL) (v*NumRates);

   arg_i1=0.0;
   arg1=0.0;

   for(ir=1;ir <= NumRates;ir++) for(jr=1;jr <=NumRates;jr++) {
     term = Couplings[ir][jr]*(RatesPres[ir]-RatesPrev[ir]) * 
                              (RatesPres[jr]-RatesPrev[jr]);
     arg_i1 += term;
     arg1 += term;
   } /* terms independent of shift */

   /* terms linear in shift */
   term = 0.0;
   for(jr=1;jr <=NumRates;jr++) {
     term += (RatesPres[jr]-RatesPrev[jr])*Couplings[TrialIndex][jr];
   }
   term *= 2.*shift;
   arg1 += term;

   /* term quadratic in shift */
   arg1 += shift*shift*Couplings[TrialIndex][TrialIndex];

   arg_i1 /= v2;
   arg1 /= v2;

   prob_fi1= (1. + arg_i1 )/(1. + arg1);
   arg = v1*log(prob_fi1);
   arg = (arg > MAXARG) ? MAXARG:arg;
   arg = (arg < MINARG) ? MINARG:arg;

   prob_fi = exp(arg);

   return prob_fi;

}/* end of generic Student (n=3) with translation symmetry */


#endif
